/* Copyright (c) 1994-95 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * Author: Satish Chandra, 1994
 * Derived from the shared memory version by Satish Chandra and
 * Steve Dirkse (University of Wisconsin Madison)
 */

#include <cm/cmmd/io.h>
#include <cm/cmmd.h>
#include <stdio.h>
#include <fcntl.h>
#include <math.h>

/******************************************************************
* Global Data                                                     *
******************************************************************/

char datafile[100] = "Data4096.01";             /* the input file */
int n = 4096;			/* default size of the system */
double contol = 1.0e-15;		/* convergence tolerance */
int kseed = 98765;		/* random number seed */
double sparsity = 0.01;		/* matrix sparsity */
double xsparsity = 0.50;	        /* solution x sparsity */
int iterlim = 400;		/* iteration limit */
int nprocs = 32;		/* number of processors, read from command line */
int nsweeps = 1;
int nsweepslo = 5;		/* lower number of sweeps */
int nsweepshi = 5;		/* higher number of sweeps */
double Gamma = 0.25;		/* relaxation parameter */
int lcp = 1;			/* solve equation or LCP */
int sync = 1;			/* if == 1, synchronous */
int verbose = 0;		/* output debugging stuff, if == 1 */
int iterlog = 1;		/* log iteration stats to terminal */
int output = 0;			/* write out the result to terminal */

int K;				/* n/nprocs */
int mynum;			/* this processor's number */
int startelem;                  /* which non zero element is the first in
				   this procs set of rows */

int do_work();
int do_time_steps();
int initialize ();
int read_data ();


double djunk[200000];           /* garbage space */
int *ijunk = (int *) djunk;      

int    *lr;		        /* n+1 row pointers */
double *qq;			/* n elements */
double *ox;			/* n elements */
double *solx;			/* n elements */
double *localx;
double *localq;
double *aa;		        /* big enough to hold nonz elements of K rows*/
double *aa_o;                   /* offsetted aa */

double global_norm;
double actual_dev;
double sol_dev;
double norm;
double local_actual_dev;
double local_sol_dev;

int *lc;			/* big enough to hold nonz elements of K rows*/
int *lc_o;                      /* offsetted lc */

/*
#define A(x)      aa[x - (startelem-1)]
#define icol(x)   lc[x - (startelem-1)]
*/
#define A(x)      aa_o[x]
#define icol(x)   lc_o[x]

#define irow(x)   lr[x]
#define q(x)      qq[x]
#define xbar(x)   ox[x]
#define xsol(x)   solx[x]
#define x(y)      localx[y]
#define lq(y)     localq[y]

#define min(a,b)		((a)<(b) ? (a) : (b))
#define max(a,b)		((a)>(b) ? (a) : (b))
#define abs(a)			((a)>0 ? (a) : -(a))


double SC_reduce_double(double, int);

read_data()
{
  int nel, endelem, nnzelem;
  int fd;
  int K;
  int i;

  fd = open(datafile, O_RDONLY);
  /* CMMD_set_io_mode(fd, CMMD_independent);*/

  if (fd < 0)
    {
      printf("Cannot read the data file\n");
      return;
    }

  K = n / nprocs;

  solx = (double *) malloc(sizeof(double)*n);
  qq = (double *) malloc(sizeof(double)*n);
  lr = (int *) malloc(sizeof(int)*(n+1));
  ox = (double *) malloc(sizeof(double)*n);

  if (read(fd, solx, n*sizeof(double)) != n*sizeof(double))
    goto error;
  if (read(fd, qq, n*sizeof(double)) != n*sizeof(double))
    goto error;
  if (read(fd, lr, (n+1)*sizeof(int)) != (n+1)*sizeof(int))
    goto error;

  startelem = irow(K*mynum);
  endelem = irow(K*(mynum+1));
  nnzelem = endelem - startelem;
  nel = irow(n) - 1;

  /*
  printf("PE = %d, Start = %d, End = %d, NnzElem = %d, Nel = %d\n",
	 mynum,
	 startelem,
	 endelem,
	 nnzelem,
	 nel);
  */
  /* allocate memory for lc and aa */

  lc = (int *) malloc(nnzelem * sizeof(int));
  aa = (double *) malloc(nnzelem * sizeof(double));

  aa_o = aa - (startelem - 1);
  lc_o = lc - (startelem - 1);

  /* we do not have lseek */
  {
    /*
       int *ijunk;
    
       ijunk = (int *) malloc(sizeof(int)*nel);
       */
    read(fd, ijunk, sizeof(int)*(startelem - 1));
    
    /* for (i = 0;i < (startelem-1);i++) read(fd, &ijunk, sizeof(int)); */
    if (read(fd, lc, nnzelem*sizeof(int)) != nnzelem*sizeof(int))
      goto error;
    /* skip past the rest */
    read(fd, ijunk, sizeof(int)*(nel + 1 - endelem));
    /* for (i = endelem; i <= nel;i++) read(fd, &ijunk, sizeof(int)); */
    /* free(ijunk); */
  }

  {
    /*
       double *djunk;

       djunk = (double *) malloc(sizeof(double)*nel);
       */
    read(fd, djunk, sizeof(double)*(startelem - 1));
  
    /* for (i = 0;i < (startelem-1);i++) read(fd, &djunk, sizeof(double)); */

    if (read(fd, aa, nnzelem*sizeof(double)) != nnzelem*sizeof(double))
      goto error;
    
    /* skip past the rest */
    read(fd, djunk, sizeof(double)*(nel + 1 - endelem));
    /* for (i = endelem; i <= nel;i++) read(fd, &djunk, sizeof(double)); */
    
    /* free(djunk); */
  }

  return;

 error:
  printf("Error occured while reading data\n");
  return;
}

main()
{
  int i,j,k;
  double dev;

  /* set CMMD_io modes */
#ifndef WWT
  CMMD_fset_io_mode(stdin, CMMD_sync_bc);
  CMMD_fset_io_mode(stdout, CMMD_independent);
  CMMD_fset_io_mode(stderr, CMMD_independent);
#endif

  nprocs = CMMD_partition_size();
  mynum = CMMD_self_address();

#ifndef WWT  
  if (read_inputs() != 0)
    goto error;
#endif

  K = n / nprocs;

  if (K * nprocs != n) {
    printf ("nprocs (%d) must be an integral factor of n (%d).\n", nprocs, n);
    goto error;
  }

  localx = (double *) malloc(sizeof(double) * K);
  localq = (double *) malloc(sizeof(double) * K);

  read_data();

  CMMD_sync_with_nodes();
  
  if (mynum == 0) printf ("Problem initialization completed.\n");

  if (mynum == 0)
    {
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);
    }

  do_work();

  if (mynum == 0) {
      CMMD_node_timer_stop(0);
      printf("Compute time %f seconds\n",
	     CMMD_node_timer_elapsed(0));
    }
  
  goto noerror;
  
 error:
  printf ("Error encountered: ending job\n");
  
 noerror:
}

int read_inputs()
{
  int linenum = 0;
  char
    line[100],
    keyword[100];
  char temp[100];
  double value;

  while ((gets(line)) != NULL) {
    linenum++;
/*    printf ("%3d: %s\n", linenum, line); */
    if (*line == '*' || *line == '\0') {
      continue;
    }
        
    /* The first line after any *'s should be data file name */
    sscanf (line, "%s %s", keyword, temp);
    if (!strncmp(keyword, "datafile", 8))
      {
	strcpy(datafile, temp);
	continue;
      }


    sscanf (line, "%s %lf", keyword, &value);
    if (!strncmp (keyword, "contol", 6))
      contol = value;
    else if (!strncmp (keyword, "gamma", 5))
      Gamma = value;
    else if (!strncmp (keyword, "iterlim", 7))
      iterlim = value;
    else if (!strncmp (keyword, "iterlog", 7))
      iterlog = value;
    else if (!strncmp (keyword, "kseed", 5))
      kseed = value;
    else if (!strncmp (keyword, "lcp", 3))
      lcp = value;
    else if (!strncmp (keyword, "N", 1))
      n = value;
    else if (!strncmp (keyword, "nsweepslo", 9))
      nsweepslo = value;
    else if (!strncmp (keyword, "nsweepshi", 9))
      nsweepshi = value;
    else if (!strncmp (keyword, "output", 6))
      output = value;
    else if (!strncmp (keyword, "sparsity", 8))
      sparsity = value;
    else if (!strncmp (keyword, "sync", 4))
      sync = value;
    else if (!strncmp (keyword, "verbose", 7))
      verbose = value;
    else if (!strncmp (keyword, "xsparsity", 9))
      xsparsity = value;
    else {
      printf ("Keyword '%s' in input line %d unrecognized.\n", keyword, linenum);
      return (-1);
    }
  }
  if (nsweepslo < 0)
    nsweepslo = 0;		/* will be bumped to one in loop */
  if (nsweepshi < nsweepslo)
    nsweepshi = nsweepslo;

  if (mynum == 0) {
    printf ("\nn\tkseed\tsparsity\txsparsity\n");
    printf ("%d\t%d\t%g\t\t%g\n\n",
	    n, kseed, sparsity, xsparsity);
    printf ("nprocs\tnsweepslo\tnsweepshi\titerlim\n");
    printf ("%d\t%d\t\t%d\t\t%d\n\n",
	    nprocs, nsweepslo, nsweepshi, iterlim);
    printf ("gamma\tsync\tlcp\titerlog\tverbose\toutput\tcontol\n");
    printf ("%.4g\t%s\t%s\t%s\t%s\t%s\t%g\n\n",
	    Gamma, (sync) ? "yes" : "no", (lcp) ? "yes" : "no",
	    (iterlog) ? "yes" : "no", (verbose) ? "yes" : "no",
	    (output) ? "yes" : "no", contol);
    printf ("Datafile = %s\n", datafile);
  }

  return (0);
}

void verify_sol (double *xbar, double *xsol, int *argdev, double *dev, int n)
{
  int i;
  double x, dt;
  double max_dev;

  *dev = 0;
  *argdev = 0;
  for (i = K * mynum;  i < (mynum + 1)*K;  i++) {
    dt = fabs(xsol[i] - xbar[i]);
    if (dt > *dev) {
      *argdev = i;
      *dev = dt;
    }
  }

#ifndef WWT
  /* max_dev = CMMD_reduce_double(*dev, CMMD_combiner_dmax); */
  max_dev = SC_reduce_double(*dev, CMMD_combiner_dmax);
#else
  max_dev = SC_reduce_double(*dev, CMMD_combiner_dmax);
#endif

  if (*dev == max_dev)
    {
      printf ("\n  maximum deviation: %g\n", *dev);
      printf ("          component: %d\n\n", *argdev);
    }
}

do_work()
{
  int j, k;
  int i;
  double dev;

#ifndef AM_PUT
  initialize_comm();
#endif

  for (k = nsweepslo; k <= nsweepshi; k++) {

    /* re-initialize x's to 0 */
    for (j = 0;  j < K;  j++) {
      x(j) = 0;
    }

    if (k < 1)
      nsweeps = 1;
    else
      nsweeps = k;

    /* first run through will have a cold cache, so
       we allow for a double run with nsweeps == 1
       */

    if (mynum == 0) printf("nsweeps = %d\n", nsweeps);

    /* inspect(); */

    do_time_steps();

    verify_sol (&xbar(0), &xsol(0), &i, &dev, n);

  }
}

#ifdef SHOW

void print_data ()
{
  int i, j, l;

  printf ("n = %d\n", n);
  for (i = 0;  i <= n;  i++) {
    printf ("rowptr[%3d]: %6d\n", i, irow(i));
  }
  i = 1;
  for (l = 0;  l < irow(n) - 1;  l++) {
    if (irow(i) <= l+1)
      i++;
    printf ("A[%3d,%3d]:  %g\n", i, icol(l), A(l));
  }
  for (i = 0;  i < n;  i++) {
    printf ("   q[%3d]: % g\n", i, q(i));
  }
  for (i = 0;  i < n;  i++) {
    printf ("xsol[%3d]: % g\n", i, xsol(i));
  }
}

show()
{
  int i;

  printf ("SOR parameter gamma: %g\n\n", Gamma);
  for (i = 0;  i < n;  i++) {
    printf("x[%3d] = %g\n", i, xbar(i));
  }
}

#endif

void do_norm_step ()
{
  int i, j, l;
  double stemp, atemp,
  sdev, adev;

  for (i = mynum * K;  i < (mynum + 1) * K;  i++) {
    atemp = stemp = 0;
    l = irow(i) - 1;
    while (l < irow(i+1) - 1) {
      j = icol(l) - 1;
      atemp += A(l) * xbar(j);
      stemp += A(l) * xsol(j);
      l++;
    }

    atemp += q(i);
    stemp += q(i);
  
    if (xbar(i) < 0)
      printf ("Error: xbar[%d] = %g\n", i, xbar(i));
  
    adev = min (xbar(i), abs(atemp));
    sdev = min (xsol(i), abs(stemp));
    local_actual_dev = max (local_actual_dev, adev);
    local_sol_dev = max (local_sol_dev, sdev);    
  }
}

void do_gauss_sweep ()
{
  int i, j, l,
  diag;
  double temp,
  denom;
  int k = 0;

  if (mynum == 0 && verbose)
    printf("Entering Gauss Sweep\n");
  for (i = mynum*K;  i < (mynum+1)*K;  i++) {
    temp = 0;
    l = irow(i) - 1;
    diag = -1;
    while (l < irow(i+1) - 1) {
      j = icol(l) - 1;
      if (j == i)
	diag = l;
      else 
	temp += A(l) * xbar(j);
      l++;
    }
    temp += q(i);

    if (diag < 0) {
      printf ("Error: no diagonal in row %d\n", i);
      return;
    }

    if (lcp) {			/* solve lcp */
      temp += A(diag) * xbar(i);
      xbar(i) = max(0, xbar(i) - Gamma*(temp/A(diag)));
    }
    else {			/* solve equation */
				/* over-relaxation step */
      xbar(i) = (1 - Gamma) * xbar(i) - Gamma * temp / A(diag);
    }
#ifdef AM_PUT
    propagate_updates(xbar(i), &(xbar(i)));
#endif
  }
}

int wwt_get_miss_time();

do_time_steps()
{
  int step_counter = 0;
  int i,j,k, l;
  double dt;
  double temp;

  if (iterlog && mynum == 0) {
    printf ("  step          norm\n");
    printf ("  ----       ----------\n");
  }

  while (step_counter++ < iterlim) {
    
    /* copy old solution */
    for (i = 0;  i < K;  i++) {
      x(i) = xbar(i + K * mynum);
    }
    
    for (i = 0;  i < nsweeps;  i++) {
      do_gauss_sweep();
#ifndef AM_PUT
      propagate_updates();
#endif
    }
    /* CMMD_sync_with_nodes(); */
    norm = 0;
    for (i = 0;  i < K;  i++) {
      dt = abs(xbar(i + K*mynum) - x(i));
      norm = max(dt,norm);
    }

    wait_for_updates();
    /* do the reduction here */

#ifndef WWT
    /* norm = CMMD_reduce_double(norm, CMMD_combiner_dmax); */
    norm = SC_reduce_double(norm, CMMD_combiner_dmax);
#else
    norm = SC_reduce_double(norm, CMMD_combiner_dmax);
#endif
    
    if (mynum == 0) global_norm = norm;

    if (iterlog && mynum == 0) {
      printf ("  %4d    %13.6e\n", step_counter, global_norm);
    }
    if (norm < contol) break;

  }

  do_norm_step ();

  /* do a reduction on local_actual_dev and local_sol_dev to get
     actual_dev and sol_dev */

#ifndef WWT
  /*
  actual_dev = CMMD_reduce_double(local_actual_dev, CMMD_combiner_dmax);
  sol_dev = CMMD_reduce_double(local_sol_dev, CMMD_combiner_dmax);
  */
  actual_dev = SC_reduce_double(local_actual_dev, CMMD_combiner_dmax);
  sol_dev = SC_reduce_double(local_sol_dev, CMMD_combiner_dmax);
#else
  actual_dev = SC_reduce_double(local_actual_dev, CMMD_combiner_dmax);
  sol_dev = SC_reduce_double(local_sol_dev, CMMD_combiner_dmax);
#endif

  if (mynum == 0) {
    if (step_counter > iterlim)
      printf ("\n ** exiting loop -- Iteration limit %d exceeded. **\n");
    else {
      printf ("\n ** exiting loop -- norm: %g\n", global_norm);
      printf ("         time steps: %d\n", step_counter);
      printf ("     predicted deviation: %g\n", sol_dev);
      printf ("        actual deviation: %g\n", actual_dev);
    }
  }
}

/* stuff related to update communication */

#ifdef AM_PUT

volatile int update_counter = 0;

void
put_handler(double *address, int v1, int v2)
{
  *((volatile int *)&update_counter) = update_counter + 1;
  *((int *)address) = v1;
  *((int *)address+1) = v2;
}

propagate_updates(double value, double *address)
{
  int i;
  int v1, v2;

  if (CMAML_interrupt_status() == FALSE) 
    printf("Wrong state.\n");

  v1 = *((int *)&value);
  v2 = *((int *)&value + 1);

  for (i = 0; i < nprocs; i++)
    if (i != mynum)
      {
	CMAML_rpc(i, (void *) &put_handler, 
		  (int) address, v1, v2, 0);
      }
}

wait_for_updates()
{
  int old_status;
  old_status = CMAML_disable_interrupts();
  while (*((volatile int *)&update_counter) < (nsweeps * (n - K)))
    {
      CMAML_dispatch();
      /* printf("PE = %d, UC = %d\n",mynum, update_counter); */
    }
  update_counter = 0;
  if (old_status) CMAML_enable_interrupts();
}

#else

#include "eventlog.h"

#define UPDATE_RPORT  75  /* aren't we having too many adhoc rports ? */


volatile int update_counter = 0;
volatile int last_reading = 0;

void
update_handler(int rport, int dummy2)
{
  update_counter++;
  /* reset byte counter */
  CMAML_set_rport_byte_counter(UPDATE_RPORT, nsweeps * (n - K) * sizeof(double));
}

initialize_comm()
{
  int t;
  int old_status;

  old_status = CMAML_disable_interrupts();

  if ((t = CMAML_allocate_this_rport(UPDATE_RPORT)) < 0)
    {
      printf("Could not allocate the desired Rport. Sorry.\n");
      exit();
    }
  CMAML_set_rport_addr(UPDATE_RPORT, &(xbar(0)));
  CMAML_set_rport_byte_counter(UPDATE_RPORT, nsweeps * (n - K) * sizeof(double));
  /* CMAML_set_rport_handler(UPDATE_RPORT, &update_handler); */
  CMAML_set_rport_handler(UPDATE_RPORT, NULL);
  CMAML_set_rport_handler_arg2(UPDATE_RPORT, NULL);
  CMAML_set_rport_total_nbytes(UPDATE_RPORT, nsweeps * (n - K) * sizeof(double));

  wwt_log_ev(EV_OPEN_RECEIVE_CHANNEL,sizeof(char), 
	     sizeof(char),nsweeps * (n - K) * sizeof(double));
  wwt_log_ev(EV_NOTIFY_CH_RCD,UPDATE_RPORT, 0, 0);

  if (old_status) CMAML_enable_interrupts();

  CMMD_sync_with_nodes();
}

propagate_updates()
{
  int i;
  int t;
  int old_status;
#ifdef WWT
  wwt_log_ev(EV_ENTER_BC, 0, 0, 0);
#endif
  old_status = CMAML_disable_interrupts();

  /*
     printf("PE %d, byte counter = %d\n", mynum, 
     CMAML_get_rport_byte_counter(UPDATE_RPORT));
     */

  t = CMAML_get_rport_byte_counter(UPDATE_RPORT);
  update_counter += (last_reading - t);
  CMAML_set_rport_byte_counter(UPDATE_RPORT, nsweeps * (n - K) * sizeof(double));
  last_reading = t;
    
  CMAML_dispatch();
  
  for (i = 0; i < nprocs; i++)
    if (((i + mynum + 1)%nprocs) != mynum)
      {
	t = CMAML_scopy(((i + mynum + 1)%nprocs), UPDATE_RPORT, 0, 
			mynum * K * sizeof(double),
			&(xbar(mynum * K)), K * sizeof(double), 
			NULL, NULL);
	CMAML_dispatch();
	if (t < 0) printf("scopy did not work\n");
      }
  
  if (old_status) CMAML_enable_interrupts();
#ifdef WWT
  wwt_log_ev(EV_EXIT_BC, 0, 0, 0);
#endif
}

wait_for_updates()
{
  int old_status;
#ifdef WWT
  wwt_log_ev(EV_ENTER_BC, 0, 0, 0);
#endif
  old_status = CMAML_disable_interrupts();
  /*
  while ((*(volatile int *) &update_counter) == 0) 
    {
      CMAML_dispatch();
    }
  update_counter = 0;
  */

  while (update_counter < nsweeps * (n - K) * sizeof(double))
    {
      int t;
      
      t = CMAML_get_rport_byte_counter(UPDATE_RPORT);
      update_counter += (last_reading - t);
      CMAML_set_rport_byte_counter(UPDATE_RPORT, nsweeps * (n - K) * sizeof(double));
      last_reading = t;
      CMAML_dispatch();
    }
  
  update_counter = 0;
  last_reading = 0;
  
  if (old_status) CMAML_enable_interrupts();
#ifdef WWT
  wwt_log_ev(EV_EXIT_BC, 0, 0, 0);
#endif
}

#endif AM_PUT
