/* Copyright (c) 1994-95 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 */
/*---------------------------------*/
/* geometry.c                      */
/*---------------------------------*/
/* module to be used by            */
/* nspher.c                        */
/*---------------------------------*/
/* programmed by Frank Traenkle    */
/*---------------------------------*/
/* creates boundary elements for   */
/* sphere of radius 1              */ 
/*---------------------------------*/

/*----------------------------*/
/* INCLUDEs                   */
/*----------------------------*/ 
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "my.h"
#include "geometry.h"

/*----------------------------*/
/* DEFINEs                    */
/*----------------------------*/
#define M_MAX 320

/*----------------------------*/
/* PROTOTYPEs                 */
/*----------------------------*/
static void CreateTetrahedron(void);
static void CreateOctahedron(void); 
static void CreateIcosahedron(void);
static void Tesselate(void);
static void Bisector(USHORT usK, USHORT usI, USHORT usJ); 
static void NewElements(USHORT usK);
static void AddElement(USHORT usI, USHORT usJ, USHORT usK);
static void RenameElement(USHORT usK, USHORT usL, USHORT usM,
                          USHORT usN);

/*----------------------------*/
/* VARIABLEs                  */
/*----------------------------*/
static FPTYPE (*afpNode)[3];
static USHORT (*ausElement)[3];
static USHORT usLastElement;
static USHORT usLastNode;
static USHORT ausNewNode[3];

/*----------------------------*/
/* FUNCTION CreateGeometry    */
/*----------------------------*/
extern USHORT CreateGeometry(USHORT usMstart, USHORT usNtess)
{
  USHORT usI;
  USHORT usM;

  /* check if number of elements is too high */
  usM = usMstart;
  for(usI = 0; usI < usNtess; usI++)
  {
    usM *= 4;
  }
  if (usM > M_MAX)
  {
    /* not enough memory for this order of tesselation */
    return 0;
  }

  /* allocate memory */
  afpNode = (FPTYPE (*)[])malloc(usM * sizeof(*afpNode));
  ausElement = (USHORT (*)[])malloc(usM * sizeof(*ausElement));

  /* create starting polyhedron */ 
  switch(usMstart)
  {
    case 4:
      CreateTetrahedron();
      break;
    case 8:
      CreateOctahedron();
      break;
    case 20:
      CreateIcosahedron();
      break;
    default:
      /* usMstart is not valid */
      return 0;
  }

  /* tesselate usNtess times */
  for(usI = 0; usI < usNtess; usI++)
  {
    Tesselate();
  }
  
  return usM;
}

/*----------------------------*/
/* FUNCTION CloseGeometry     */
/*----------------------------*/
extern void CloseGeometry(void)
{
  free(afpNode);
  free(ausElement);
}  
 
/*----------------------------*/
/* FUNCTION CreateTetrahedron */
/*----------------------------*/
static void CreateTetrahedron(void) 
{
  FPTYPE fpTheta;

  fpTheta = PI - acos(1. / 3.);

  afpNode[0][0] = 0.0;
  afpNode[0][1] = 0.0;
  afpNode[0][2] = 1.0;
                       
  afpNode[1][0] = sin(fpTheta) * sin(0.0);
  afpNode[1][1] = sin(fpTheta) * cos(0.0);
  afpNode[1][2] = cos(fpTheta);

  afpNode[2][0] = sin(fpTheta) * sin(2. / 3. * PI);
  afpNode[2][1] = sin(fpTheta) * cos(2. / 3. * PI);
  afpNode[2][2] = cos(fpTheta);

  afpNode[3][0] = sin(fpTheta) * sin(4. / 3. * PI);
  afpNode[3][1] = sin(fpTheta) * cos(4. / 3. * PI);
  afpNode[3][2] = cos(fpTheta);             

  ausElement[0][0] = 0;
  ausElement[0][1] = 1;       
  ausElement[0][2] = 2;       
  ausElement[1][0] = 0;       
  ausElement[1][1] = 1;       
  ausElement[1][2] = 3;       
  ausElement[2][0] = 0;       
  ausElement[2][1] = 2;       
  ausElement[2][2] = 3;       
  ausElement[3][0] = 1;       
  ausElement[3][1] = 2;       
  ausElement[3][2] = 3;

  usLastElement = 4;
  usLastNode = 4;
}

/*----------------------------*/
/* FUNCTION CreateOctahedron  */
/*----------------------------*/
static void CreateOctahedron(void) 
{
  afpNode[0][0] = 0.0;
  afpNode[0][1] = 0.0;
  afpNode[0][2] = 1.0;
                 
  afpNode[1][0] = 1.0;
  afpNode[1][1] = 0.0;
  afpNode[1][2] = 0.0;
             
  afpNode[2][0] = 0.0;
  afpNode[2][1] = 1.0;
  afpNode[2][2] = 0.0;

  afpNode[3][0] = -1.0;
  afpNode[3][1] = 0.0;
  afpNode[3][2] = 0.0;

  afpNode[4][0] = 0.0;
  afpNode[4][1] = -1.0;
  afpNode[4][2] = 0.0;

  afpNode[5][0] = 0.0;
  afpNode[5][1] = 0.0;
  afpNode[5][2] = -1.0;

  ausElement[0][0] = 0;
  ausElement[0][1] = 1;       
  ausElement[0][2] = 2;       
  ausElement[1][0] = 0;       
  ausElement[1][1] = 2;       
  ausElement[1][2] = 3;       
  ausElement[2][0] = 0;       
  ausElement[2][1] = 3;       
  ausElement[2][2] = 4;       
  ausElement[3][0] = 0;       
  ausElement[3][1] = 4;       
  ausElement[3][2] = 1;
  ausElement[4][0] = 5;       
  ausElement[4][1] = 1;       
  ausElement[4][2] = 2;
  ausElement[5][0] = 5;       
  ausElement[5][1] = 2;       
  ausElement[5][2] = 3;
  ausElement[6][0] = 5;       
  ausElement[6][1] = 3;       
  ausElement[6][2] = 4;
  ausElement[7][0] = 5;       
  ausElement[7][1] = 4;       
  ausElement[7][2] = 1;
              
  usLastElement = 8;
  usLastNode = 6;
}

/*----------------------------*/
/* FUNCTION CreateIcosahedron */
/*----------------------------*/
static void CreateIcosahedron(void)
{
  FPTYPE fpDihedr = 138.1897222 / 180. * PI;
  FPTYPE fpSide = 4. / sqrt(2. * (5. + sqrt(5.)));
  FPTYPE fpHeight = fpSide * .5 * sqrt(3.);
  FPTYPE fpChord = sqrt(2. * fpHeight * fpHeight * 
                   (1. - cos(fpDihedr)));
  FPTYPE fpTheta1 = PI - acos((2. - fpChord * fpChord) * .5);
  FPTYPE fpTheta2 = PI - fpTheta1;

  afpNode[0][0] = 0.0;
  afpNode[0][1] = 0.0;
  afpNode[0][2] = 1.0;

  afpNode[1][0] = sin(fpTheta1) * sin(0.0);
  afpNode[1][1] = sin(fpTheta1) * cos(0.0);
  afpNode[1][2] = cos(fpTheta1);

  afpNode[2][0] = sin(fpTheta1) * sin(2./5.*PI);
  afpNode[2][1] = sin(fpTheta1) * cos(2./5.*PI);
  afpNode[2][2] = cos(fpTheta1); 

  afpNode[3][0] = sin(fpTheta1) * sin(4./5.*PI);
  afpNode[3][1] = sin(fpTheta1) * cos(4./5.*PI);
  afpNode[3][2] = cos(fpTheta1);              

  afpNode[4][0] = sin(fpTheta1) * sin(6./5.*PI);
  afpNode[4][1] = sin(fpTheta1) * cos(6./5.*PI);
  afpNode[4][2] = cos(fpTheta1);              

  afpNode[5][0] = sin(fpTheta1) * sin(8./5.*PI);
  afpNode[5][1] = sin(fpTheta1) * cos(8./5.*PI);
  afpNode[5][2] = cos(fpTheta1);

  afpNode[6][0] = 0.0;
  afpNode[6][1] = 0.0;
  afpNode[6][2] = -1.0;

  afpNode[7][0] = sin(fpTheta2) * sin(1./5.*PI);
  afpNode[7][1] = sin(fpTheta2) * cos(1./5.*PI);
  afpNode[7][2] = cos(fpTheta2);              

  afpNode[8][0] = sin(fpTheta2) * sin(3./5.*PI);
  afpNode[8][1] = sin(fpTheta2) * cos(3./5.*PI);
  afpNode[8][2] = cos(fpTheta2);              

  afpNode[9][0] = sin(fpTheta2) * sin(PI);
  afpNode[9][1] = sin(fpTheta2) * cos(PI);
  afpNode[9][2] = cos(fpTheta2);

  afpNode[10][0] = sin(fpTheta2) * sin(7./5.*PI);
  afpNode[10][1] = sin(fpTheta2) * cos(7./5.*PI);
  afpNode[10][2] = cos(fpTheta2); 

  afpNode[11][0] = sin(fpTheta2) * sin(9./5.*PI);
  afpNode[11][1] = sin(fpTheta2) * cos(9./5.*PI);
  afpNode[11][2] = cos(fpTheta2);              

  ausElement[0][0] = 0;
  ausElement[0][1] = 1;       
  ausElement[0][2] = 2;       

  ausElement[1][0] = 0;       
  ausElement[1][1] = 2;       
  ausElement[1][2] = 3;       

  ausElement[2][0] = 0;       
  ausElement[2][1] = 3;       
  ausElement[2][2] = 4;       

  ausElement[3][0] = 0;       
  ausElement[3][1] = 4;       
  ausElement[3][2] = 5;

  ausElement[4][0] = 0;
  ausElement[4][1] = 5;      
  ausElement[4][2] = 1;       

  ausElement[5][0] = 1;       
  ausElement[5][1] = 2;       
  ausElement[5][2] = 7;       

  ausElement[6][0] = 2;       
  ausElement[6][1] = 3;       
  ausElement[6][2] = 8;       

  ausElement[7][0] = 3;       
  ausElement[7][1] = 4;       
  ausElement[7][2] = 9;

  ausElement[8][0] = 4;  
  ausElement[8][1] = 5;       
  ausElement[8][2] = 10;      

  ausElement[9][0] = 5;       
  ausElement[9][1] = 1;       
  ausElement[9][2] = 11;      

  ausElement[10][0] = 1;       
  ausElement[10][1] = 7;       
  ausElement[10][2] = 11;

  ausElement[11][0] = 2;       
  ausElement[11][1] = 7;       
  ausElement[11][2] = 8; 
                                     
  ausElement[12][0] = 3;       
  ausElement[12][1] = 8;       
  ausElement[12][2] = 9;

  ausElement[13][0] = 4;       
  ausElement[13][1] = 9;      
  ausElement[13][2] = 10;

  ausElement[14][0] = 5;       
  ausElement[14][1] = 10;      
  ausElement[14][2] = 11;

  ausElement[15][0] = 6;       
  ausElement[15][1] = 7;       
  ausElement[15][2] = 8;       

  ausElement[16][0] = 6;       
  ausElement[16][1] = 8;       
  ausElement[16][2] = 9;
            
  ausElement[17][0] = 6;       
  ausElement[17][1] = 9;      
  ausElement[17][2] = 10;

  ausElement[18][0] = 6;       
  ausElement[18][1] = 10;      
  ausElement[18][2] = 11;

  ausElement[19][0] = 6;       
  ausElement[19][1] = 11;      
  ausElement[19][2] = 7; 

  usLastElement = 20;
  usLastNode = 12;
}

/*----------------------------*/
/* FUNCTION Tesselate         */
/*----------------------------*/
static void Tesselate(void) 
{
  USHORT usLocalLastElement;
  USHORT usK;

  usLocalLastElement = usLastElement;
  for(usK = 0; usK < usLocalLastElement; usK++)
  {
    Bisector(usK, 0, 1);
    Bisector(usK, 1, 2);
    Bisector(usK, 2, 0);
    NewElements(usK);
  } 
}

/*----------------------------*/
/* FUNCTION Bisector          */
/*----------------------------*/
static void Bisector(USHORT usK, USHORT usI, USHORT usJ) 
{
  USHORT usEndX1, usEndX2;
  FPTYPE fpNorm;
  FPTYPE fpE00, fpE01, fpE02, fpE10, fpE11, fpE12;
  FPTYPE afpX[3];

  usEndX1 = ausElement[usK][usI];
  usEndX2 = ausElement[usK][usJ];

  fpE00 = afpNode[usEndX1][0];
  fpE01 = afpNode[usEndX1][1];
  fpE02 = afpNode[usEndX1][2];
  fpE10 = afpNode[usEndX2][0];
  fpE11 = afpNode[usEndX2][1];
  fpE12 = afpNode[usEndX2][2]; 

  afpX[0] = fpE00 + fpE10;
  afpX[1] = fpE01 + fpE11;
  afpX[2] = fpE02 + fpE12; 

  /* project new node onto surface */
  fpNorm = sqrt(afpX[0]*afpX[0] + afpX[1]*afpX[1] + afpX[2]*afpX[2]);
  afpX[0] /= fpNorm;
  afpX[1] /= fpNorm;
  afpX[2] /= fpNorm;
   
  afpNode[usLastNode][0] = afpX[0];
  afpNode[usLastNode][1] = afpX[1];
  afpNode[usLastNode][2] = afpX[2];
  ausNewNode[usI] = usLastNode;

  usLastNode++;
}
                       
/*----------------------------*/
/* FUNCTION NewElements       */
/*----------------------------*/
static void NewElements(USHORT usK)
{
  USHORT usNewNodeX1 = ausNewNode[0]; 
  USHORT usNewNodeX2 = ausNewNode[1];
  USHORT usNewNodeX3 = ausNewNode[2];

  USHORT usOldNodeX1 = ausElement[usK][0];
  USHORT usOldNodeX2 = ausElement[usK][1];
  USHORT usOldNodeX3 = ausElement[usK][2];

  AddElement(usOldNodeX1, usNewNodeX1, usNewNodeX3);
  AddElement(usOldNodeX2, usNewNodeX1, usNewNodeX2); 
  AddElement(usOldNodeX3, usNewNodeX2, usNewNodeX3); 

  RenameElement(usK, usNewNodeX1, usNewNodeX2, usNewNodeX3);
}

/*----------------------------*/
/* FUNCTION AddElement        */
/*----------------------------*/
static void AddElement(USHORT usI, USHORT usJ, USHORT usK)
{
  ausElement[usLastElement][0] = usI;
  ausElement[usLastElement][1] = usJ;
  ausElement[usLastElement][2] = usK;
  usLastElement++;
}

/*----------------------------*/
/* FUNCTION RenameElement     */
/*----------------------------*/
static void RenameElement(USHORT usK, USHORT usL, USHORT usM,
                          USHORT usN)
{
  ausElement[usK][0] = usL;
  ausElement[usK][1] = usM;
  ausElement[usK][2] = usN;
}

/*----------------------------*/
/* FUNCTION GetElementAll     */
/*----------------------------*/
extern void GetElementAll(USHORT usK, PGEO pgeo)
{
  USHORT usNode1 = ausElement[usK][0];
  USHORT usNode2 = ausElement[usK][1];
  USHORT usNode3 = ausElement[usK][2];
  FPTYPE afpP1[3], afpP2[3], afpP3[3];
  FPTYPE afpU1[3], afpU2[3];
  FPTYPE afpX[3];
  FPTYPE afpN[3], fpDot, fpNorm; 

  /* vertices */
  pgeo->afpP1[0] = afpP1[0] = afpNode[usNode1][0];
  pgeo->afpP1[1] = afpP1[1] = afpNode[usNode1][1];
  pgeo->afpP1[2] = afpP1[2] = afpNode[usNode1][2];
  pgeo->afpP2[0] = afpP2[0] = afpNode[usNode2][0];
  pgeo->afpP2[1] = afpP2[1] = afpNode[usNode2][1];
  pgeo->afpP2[2] = afpP2[2] = afpNode[usNode2][2];
  pgeo->afpP3[0] = afpP3[0] = afpNode[usNode3][0];
  pgeo->afpP3[1] = afpP3[1] = afpNode[usNode3][1];
  pgeo->afpP3[2] = afpP3[2] = afpNode[usNode3][2];
  
  /* vectors of edges */
  pgeo->afpU1[0] = afpU1[0] = afpP1[0] - afpP2[0];
  pgeo->afpU1[1] = afpU1[1] = afpP1[1] - afpP2[1];
  pgeo->afpU1[2] = afpU1[2] = afpP1[2] - afpP2[2];
  pgeo->afpU2[0] = afpU2[0] = afpP3[0] - afpP2[0];
  pgeo->afpU2[1] = afpU2[1] = afpP3[1] - afpP2[1];
  pgeo->afpU2[2] = afpU2[2] = afpP3[2] - afpP2[2];

  /* centroid */
  pgeo->afpX[0] = afpX[0] = (afpP1[0] + afpP2[0] + afpP3[0]) / 3.;
  pgeo->afpX[1] = afpX[1] = (afpP1[1] + afpP2[1] + afpP3[1]) / 3.;
  pgeo->afpX[2] = afpX[2] = (afpP1[2] + afpP2[2] + afpP3[2]) / 3.;

  /* jacobian, area, normal */
  afpN[0] = afpU1[1]*afpU2[2] - afpU1[2]*afpU2[1];
  afpN[1] = afpU1[2]*afpU2[0] - afpU1[0]*afpU2[2];
  afpN[2] = afpU1[0]*afpU2[1] - afpU1[1]*afpU2[0];
  pgeo->fpJacob = fpNorm = sqrt(afpN[0]*afpN[0] + afpN[1]*afpN[1] + 
                                afpN[2]*afpN[2]);
  pgeo->fpDA = fpNorm * .5;
  afpN[0] /= fpNorm;
  afpN[1] /= fpNorm;
  afpN[2] /= fpNorm;
  fpDot = afpX[0]*afpN[0] + afpX[1]*afpN[1] + afpX[2]*afpN[2];
  if (fpDot < 0.)
  {
    pgeo->afpN[0] = -afpN[0];
    pgeo->afpN[1] = -afpN[1];
    pgeo->afpN[2] = -afpN[2];
  } 
  else
  {
    pgeo->afpN[0] = afpN[0];
    pgeo->afpN[1] = afpN[1];
    pgeo->afpN[2] = afpN[2];
  } 
}
